@@ -1,6 +1,6 @@
-/*         ______   ___    ___ 
+/*         ______   ___    ___
  *        /\  _  \ /\_ \  /\_ \
- *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
+ *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
  *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
  *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
  *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
@@ -38,22 +38,25 @@ struct system_data_t {
    jclass illegal_argument_exception_class;
    jclass apk_stream_class;
    jclass image_loader_class;
-   
+   jclass clipboard_class;
+   jclass apk_fs_class;
+
    ALLEGRO_SYSTEM_ANDROID *system;
    ALLEGRO_MUTEX *mutex;
    ALLEGRO_COND *cond;
    ALLEGRO_THREAD *trampoline;
    bool trampoline_running;
-   
+
    ALLEGRO_USTR *user_lib_name;
    ALLEGRO_USTR *resources_dir;
    ALLEGRO_USTR *data_dir;
    ALLEGRO_USTR *apk_path;
    ALLEGRO_USTR *model;
-	
+   ALLEGRO_USTR *manufacturer;
+
    void *user_lib;
    int (*user_main)(int argc, char **argv);
-   
+
    int orientation;
 
    bool is_2_1;   /* is running on Android OS 2.1? */
@@ -91,11 +94,21 @@ jclass _al_android_image_loader_class(void)
    return system_data.image_loader_class;
 }
 
+jclass _al_android_clipboard_class(void)
+{
+   return system_data.clipboard_class;
+}
+
 jobject _al_android_activity_object()
 {
    return system_data.activity_object;
 }
 
+jclass _al_android_apk_fs_class(void)
+{
+   return system_data.apk_fs_class;
+}
+
 static void finish_activity(JNIEnv *env);
 
 static bool already_cleaned_up = false;
@@ -171,29 +184,35 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
    if(system_data.system) {
       return true;
    }
-   
+
    pthread_t self = pthread_self();
    ALLEGRO_DEBUG("pthread_self:%p", (void*)self);
    ALLEGRO_DEBUG("nativeOnCreate begin");
-   
+
    memset(&system_data, 0, sizeof(system_data));
-   
+
    ALLEGRO_DEBUG("grab activity global refs");
    system_data.env             = env;
    system_data.activity_object = (*env)->NewGlobalRef(env, obj);
-   
+
    iae = (*env)->FindClass(env, "java/lang/IllegalArgumentException");
    system_data.illegal_argument_exception_class = (*env)->NewGlobalRef(env, iae);
-   
+
    aisc = (*env)->FindClass(env, ALLEGRO_ANDROID_PACKAGE_NAME_SLASH "/AllegroInputStream");
    system_data.input_stream_class = (*env)->NewGlobalRef(env, aisc);
-   
+
    asc = (*env)->FindClass(env, ALLEGRO_ANDROID_PACKAGE_NAME_SLASH "/AllegroAPKStream");
    system_data.apk_stream_class = (*env)->NewGlobalRef(env, asc);
 
    asc = (*env)->FindClass(env, ALLEGRO_ANDROID_PACKAGE_NAME_SLASH "/ImageLoader");
    system_data.image_loader_class = (*env)->NewGlobalRef(env, asc);
 
+   asc = (*env)->FindClass(env, ALLEGRO_ANDROID_PACKAGE_NAME_SLASH "/Clipboard");
+   system_data.clipboard_class = (*env)->NewGlobalRef(env, asc);
+
+   asc = (*env)->FindClass(env, ALLEGRO_ANDROID_PACKAGE_NAME_SLASH "/AllegroAPKList");
+   system_data.apk_fs_class = (*env)->NewGlobalRef(env, asc);
+
    ALLEGRO_DEBUG("create mutex and cond objects");
    system_data.mutex = al_create_mutex();
    system_data.cond  = al_create_cond();
@@ -204,23 +223,25 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
    system_data.data_dir = _jni_callStringMethod(env, system_data.activity_object, "getPubDataDir", "()Ljava/lang/String;");
    system_data.apk_path = _jni_callStringMethod(env, system_data.activity_object, "getApkPath", "()Ljava/lang/String;");
    system_data.model = _jni_callStringMethod(env, system_data.activity_object, "getModel", "()Ljava/lang/String;");
+   system_data.manufacturer = _jni_callStringMethod(env, system_data.activity_object, "getManufacturer", "()Ljava/lang/String;");
    ALLEGRO_DEBUG("resources_dir: %s", al_cstr(system_data.resources_dir));
    ALLEGRO_DEBUG("data_dir: %s", al_cstr(system_data.data_dir));
    ALLEGRO_DEBUG("apk_path: %s", al_cstr(system_data.apk_path));
    ALLEGRO_DEBUG("model: %s", al_cstr(system_data.model));
+   ALLEGRO_DEBUG("manufacturer: %s", al_cstr(system_data.manufacturer));
 
    ALLEGRO_DEBUG("creating ALLEGRO_SYSTEM_ANDROID struct");
    na_sys = system_data.system = (ALLEGRO_SYSTEM_ANDROID*)al_malloc(sizeof *na_sys);
    memset(na_sys, 0, sizeof *na_sys);
-   
+
    ALLEGRO_DEBUG("get system pointer");
    ALLEGRO_SYSTEM *sys = &na_sys->system;
    ALLEGRO_DEBUG("get system interface");
    sys->vt = _al_system_android_interface();
-   
+
    ALLEGRO_DEBUG("init display vector");
    _al_vector_init(&sys->displays, sizeof(ALLEGRO_DISPLAY_ANDROID *));
-   
+
    ALLEGRO_DEBUG("init time");
    _al_unix_init_time();
 
@@ -248,7 +269,7 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
    ALLEGRO_DEBUG("creating trampoline for app thread");
    system_data.trampoline = al_create_thread(android_app_trampoline, NULL);
    al_start_thread(system_data.trampoline);
-   
+
    ALLEGRO_DEBUG("waiting for app trampoline to signal running");
    al_lock_mutex(system_data.mutex);
    while(!system_data.trampoline_running) {
@@ -257,7 +278,7 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
    al_unlock_mutex(system_data.mutex);
 
    ALLEGRO_DEBUG("setup done. returning to dalvik.");
-   
+
    return true;
 }
 
@@ -265,26 +286,36 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
 JNI_FUNC(void, AllegroActivity, nativeOnPause, (JNIEnv *env, jobject obj))
 {
    (void)env; (void)obj;
-   
+
    ALLEGRO_DEBUG("pause activity\n");
 
    system_data.paused = true;
-   
+
    ALLEGRO_SYSTEM *sys = (void *)al_get_system_driver();
-   ASSERT(sys != NULL);
+
+   if(!system_data.system || !sys) {
+      ALLEGRO_DEBUG("no system driver");
+      return;
+   }
+
+   // ASSERT(sys != NULL);
 
    ALLEGRO_DISPLAY *display = *(ALLEGRO_DISPLAY**)_al_vector_ref(&sys->displays, 0);
 
    if (display) {
       ALLEGRO_EVENT event;
       _al_event_source_lock(&display->es);
-      
+
       if(_al_event_source_needs_to_generate_event(&display->es)) {
          event.display.type = ALLEGRO_EVENT_DISPLAY_SWITCH_OUT;
          event.display.timestamp = al_current_time();
          _al_event_source_emit_event(&display->es, &event);
       }
       _al_event_source_unlock(&display->es);
+
+      ALLEGRO_DISPLAY_ANDROID *d = (ALLEGRO_DISPLAY_ANDROID*)display;
+      if (d->created)
+         _al_android_destroy_surface(env, d, false);
    }
 }
 
@@ -292,13 +323,13 @@ JNI_FUNC(void, AllegroActivity, nativeOnResume, (JNIEnv *env, jobject obj))
 {
    ALLEGRO_SYSTEM *sys = &system_data.system->system;
    ALLEGRO_DISPLAY *d = NULL;
-   
+
    (void)obj;
-   
+
    system_data.paused = false;
-   
+
    ALLEGRO_DEBUG("resume activity");
-   
+
    if(!system_data.system || !sys) {
       ALLEGRO_DEBUG("no system driver");
       return;
@@ -308,20 +339,20 @@ JNI_FUNC(void, AllegroActivity, nativeOnResume, (JNIEnv *env, jobject obj))
       ALLEGRO_DEBUG("no display, not sending SWITCH_IN event");
       return;
    }
-   
+
    d = *(ALLEGRO_DISPLAY**)_al_vector_ref(&sys->displays, 0);
    ALLEGRO_DEBUG("got display: %p", d);
-   
+
    if(!((ALLEGRO_DISPLAY_ANDROID*)d)->created) {
       _al_android_create_surface(env, true); // request android create our surface
    }
-   
+
    ALLEGRO_DISPLAY *display = *(ALLEGRO_DISPLAY**)_al_vector_ref(&sys->displays, 0);
 
    if (display) {
       ALLEGRO_EVENT event;
       _al_event_source_lock(&display->es);
-      
+
       if(_al_event_source_needs_to_generate_event(&display->es)) {
          event.display.type = ALLEGRO_EVENT_DISPLAY_SWITCH_IN;
          event.display.timestamp = al_current_time();
@@ -354,25 +385,25 @@ JNI_FUNC(void, AllegroActivity, nativeOnDestroy, (JNIEnv *env, jobject obj))
       android_cleanup(false);
       return;
    }
-   
+
    ALLEGRO_DEBUG("destroy activity");
    if(!system_data.user_lib) {
       ALLEGRO_DEBUG("user lib not loaded.");
       return;
    }
-   
+
    system_data.user_main = NULL;
    if(dlclose(system_data.user_lib) != 0) {
       ALLEGRO_ERROR("failed to unload user lib: %s", dlerror());
       return;
    }
-   
+
    (*env)->DeleteGlobalRef(env, system_data.activity_object);
    (*env)->DeleteGlobalRef(env, system_data.illegal_argument_exception_class);
    (*env)->DeleteGlobalRef(env, system_data.input_stream_class);
-   
+
    free(system_data.system);
-   
+
    memset(&system_data, 0, sizeof(system_data));
 }
 
@@ -383,26 +414,26 @@ JNI_FUNC(void, AllegroActivity, nativeOnOrientationChange, (JNIEnv *env, jobject
    ALLEGRO_EVENT event;
 
    (void)env; (void)obj;
-   
+
    ALLEGRO_DEBUG("got orientation change!");
-   
+
    system_data.orientation = orientation;
-      
+
    if (!init) {
-         
+
       /* no display, just skip */
       if (!_al_vector_size(&sys->displays)) {
          ALLEGRO_DEBUG("no display, not sending orientation change event");
          return;
       }
-      
+
       d = *(ALLEGRO_DISPLAY**)_al_vector_ref(&sys->displays, 0);
       ASSERT(d != NULL);
-      
+
       ALLEGRO_DEBUG("locking display event source: %p %p", d, &d->es);
-      
+
       _al_event_source_lock(&d->es);
-      
+
       if(_al_event_source_needs_to_generate_event(&d->es)) {
          ALLEGRO_DEBUG("emit event");
          event.display.type = ALLEGRO_EVENT_DISPLAY_ORIENTATION;
@@ -410,13 +441,30 @@ JNI_FUNC(void, AllegroActivity, nativeOnOrientationChange, (JNIEnv *env, jobject
          event.display.orientation = orientation;
          _al_event_source_emit_event(&d->es, &event);
       }
-      
+
       ALLEGRO_DEBUG("unlocking display event source");
       _al_event_source_unlock(&d->es);
-   
+
    }
 }
 
+JNI_FUNC(void, AllegroActivity, nativeSendJoystickConfigurationEvent, (JNIEnv *env, jobject obj))
+{
+   (void)env;
+   (void)obj;
+
+   if (!al_is_joystick_installed()) {
+      return;
+   }
+
+   ALLEGRO_EVENT_SOURCE *es = al_get_joystick_event_source();
+   _al_event_source_lock(es);
+   ALLEGRO_EVENT event;
+   event.type = ALLEGRO_EVENT_JOYSTICK_CONFIGURATION;
+   _al_event_source_emit_event(es, &event);
+   _al_event_source_unlock(es);
+}
+
 static void finish_activity(JNIEnv *env)
 {
    ALLEGRO_DEBUG("pre post");
@@ -427,7 +475,7 @@ static void finish_activity(JNIEnv *env)
 static ALLEGRO_SYSTEM *android_initialize(int flags)
 {
    (void)flags;
-   
+
    ALLEGRO_DEBUG("android_initialize");
 
    /* This was stored before user main ran, to make it easy and accessible
@@ -440,14 +488,7 @@ static ALLEGRO_SYSTEM *android_initialize(int flags)
 
 static ALLEGRO_JOYSTICK_DRIVER *android_get_joystick_driver(void)
 {
-   if (strstr(al_cstr(system_data.model), "OUYA")) {
-      ALLEGRO_DEBUG("Using Linux joystick driver");
-      return &_al_joydrv_linux;
-   }
-   else {
-      ALLEGRO_DEBUG("Using Android joystick driver");
-      return &_al_android_joystick_driver;
-   }
+   return &_al_android_joystick_driver;
 }
 
 static int android_get_num_video_adapters(void)
@@ -455,6 +496,24 @@ static int android_get_num_video_adapters(void)
    return 1;
 }
 
+static bool android_get_monitor_info(int adapter, ALLEGRO_MONITOR_INFO *info)
+{
+   if (adapter >= android_get_num_video_adapters())
+      return false;
+
+   JNIEnv * env = (JNIEnv *)_al_android_get_jnienv();
+   jobject rect = _jni_callObjectMethod(env, _al_android_activity_object(), "getDisplaySize", "()Landroid/graphics/Rect;");
+
+   info->x1 = 0;
+   info->y1 = 0;
+   info->x2 = _jni_callIntMethod(env, rect, "width");
+   info->y2 = _jni_callIntMethod(env, rect, "height");
+
+   ALLEGRO_DEBUG("Monitor Info: %d:%d", info->x2, info->y2);
+
+   return true;
+}
+
 static void android_shutdown_system(void)
 {
    ALLEGRO_SYSTEM *s = al_get_system_driver();
@@ -462,10 +521,9 @@ static void android_shutdown_system(void)
    while (_al_vector_size(&s->displays) > 0) {
       ALLEGRO_DISPLAY **dptr = _al_vector_ref(&s->displays, 0);
       ALLEGRO_DISPLAY *d = *dptr;
-      _al_destroy_display_bitmaps(d);
       al_destroy_display(d);
    }
-   _al_vector_free(&s->displays);  
+   _al_vector_free(&s->displays);
 }
 
 static bool android_inhibit_screensaver(bool inhibit)
@@ -479,10 +537,10 @@ ALLEGRO_SYSTEM_INTERFACE *_al_system_android_interface()
 {
    if(android_vt)
       return android_vt;
-   
+
    android_vt = al_malloc(sizeof *android_vt);
    memset(android_vt, 0, sizeof *android_vt);
-   
+
    android_vt->initialize = android_initialize;
    android_vt->get_display_driver = _al_get_android_display_driver;
    android_vt->get_keyboard_driver = _al_get_android_keyboard_driver;
@@ -490,23 +548,24 @@ ALLEGRO_SYSTEM_INTERFACE *_al_system_android_interface()
    android_vt->get_touch_input_driver = _al_get_android_touch_input_driver;
    android_vt->get_joystick_driver = android_get_joystick_driver;
    android_vt->get_num_video_adapters = android_get_num_video_adapters;
+   android_vt->get_monitor_info = android_get_monitor_info;
    android_vt->get_path = _al_android_get_path;
    android_vt->shutdown_system = android_shutdown_system;
    android_vt->inhibit_screensaver = android_inhibit_screensaver;
-   
+
    return android_vt;
 }
 
 ALLEGRO_PATH *_al_android_get_path(int id)
 {
    ALLEGRO_PATH *path = NULL;
-   
+
    switch(id) {
       case ALLEGRO_RESOURCES_PATH:
          /* path to bundle's files */
          path = al_create_path_for_directory(al_cstr(system_data.resources_dir));
          break;
-         
+
       case ALLEGRO_TEMP_PATH:
       case ALLEGRO_USER_DATA_PATH:
       case ALLEGRO_USER_HOME_PATH:
@@ -515,18 +574,18 @@ ALLEGRO_PATH *_al_android_get_path(int id)
          /* path to sdcard */
          path = al_create_path_for_directory(al_cstr(system_data.data_dir));
          break;
-         
+
       case ALLEGRO_EXENAME_PATH:
          /* bundle path + bundle name */
-         // FIXME! 
+         // FIXME!
          path = al_create_path(al_cstr(system_data.apk_path));
          break;
-			
+
 		default:
 			path = al_create_path_for_directory("/DANGER/WILL/ROBINSON");
 			break;
    }
-   
+
    return path;
 }
 
@@ -588,7 +647,7 @@ void _al_register_system_interfaces(void)
    /* add the native activity driver */
    add = _al_vector_alloc_back(&_al_system_interfaces);
    *add = _al_system_android_interface();
-   
+
    /* TODO: add the non native activity driver */
 }
 
