@@ -38,6 +38,7 @@
 #include <sys/stat.h>
 #endif
 
+ALLEGRO_DEBUG_CHANNEL("stdio")
 
 /* forward declaration */
 const struct ALLEGRO_FILE_INTERFACE _al_file_interface_stdio;
@@ -101,6 +102,8 @@ static void *file_stdio_fopen(const char *path, const char *mode)
    FILE *fp;
    USERDATA *userdata;
 
+   ALLEGRO_DEBUG("opening %s %s\n", path, mode);
+
 #ifdef ALLEGRO_WINDOWS
    {
       wchar_t *wpath = _al_win_utf16(path);
