@@ -91,7 +91,6 @@ ALLEGRO_DEBUG_CHANNEL("opengl")
 
 
 
-#if !defined ALLEGRO_CFG_OPENGLES
 static uint32_t parse_opengl_version(const char *s)
 {
    char *p = (char *) s;
@@ -118,34 +117,33 @@ static uint32_t parse_opengl_version(const char *s)
    ALLEGRO_DEBUG("Parsed '%s' as 0x%08x\n", s, ver);
    return ver;
 }
-#endif
 
 
 
 /* Reads version info out of glGetString(GL_VERSION) */
 static uint32_t _al_ogl_version(void)
 {
-#if !defined ALLEGRO_CFG_OPENGLES
-   ALLEGRO_CONFIG *cfg;
-   const char *str;
-
-   cfg = al_get_system_config();
-   if (cfg) {
-      char const *value = al_get_config_value(cfg,
-	 "opengl", "force_opengl_version");
-      if (value) {
-         uint32_t v = parse_opengl_version(value);
-         ALLEGRO_INFO("OpenGL version forced to %d.%d.%d.%d.\n",
-            (v >> 24) & 0xff,
-            (v >> 16) & 0xff,
-            (v >> 8) & 0xff,
-            (v & 0xff));
-         return v;
-      }
+   char const *value = al_get_config_value(al_get_system_config(), "opengl",
+      "force_opengl_version");
+   if (value) {
+      uint32_t v = parse_opengl_version(value);
+      ALLEGRO_INFO("OpenGL version forced to %d.%d.%d.%d.\n",
+         (v >> 24) & 0xff,
+         (v >> 16) & 0xff,
+         (v >> 8) & 0xff,
+         (v & 0xff));
+      return v;
    }
 
+   const char *str;
+
    str = (const char *)glGetString(GL_VERSION);
    if (str) {
+      #ifdef ALLEGRO_CFG_OPENGLES
+      char *str2 = strstr(str, "ES ");
+      if (str2)
+         str = str2 + 3;
+      #endif
       return parse_opengl_version(str);
    }
    else {
@@ -154,16 +152,7 @@ static uint32_t _al_ogl_version(void)
        */
       return _ALLEGRO_OPENGL_VERSION_1_0;
    }
-#else
-   /* XXX this is asking for trouble, and should be documented */
-   const char *s = (char *)glGetString(GL_VERSION);
-   if (strstr(s, "2.0"))
-      return _ALLEGRO_OPENGL_VERSION_2_0;
-   else if (strstr(s, "1.1"))
-      return _ALLEGRO_OPENGL_VERSION_1_1;
-   else
-      return _ALLEGRO_OPENGL_VERSION_0;
-#endif
+
 }
 
 
@@ -277,6 +266,36 @@ static ALLEGRO_OGL_EXT_API *create_extension_api_table(void)
 
 
 
+typedef void (*VOID_FPTR)(void);
+/* GCC 4.8.2 and possibly others are really slow at optimizing the 100's of the
+ * if statements in the load_extensions function below, so we extract them to
+ * this function.
+ */
+static VOID_FPTR load_extension(const char* name)
+{
+   VOID_FPTR fptr = NULL;
+#ifdef ALLEGRO_WINDOWS
+   fptr = (VOID_FPTR)wglGetProcAddress(name);
+#elif defined ALLEGRO_UNIX
+   fptr = (VOID_FPTR)alXGetProcAddress((const GLubyte*)name);
+#elif defined ALLEGRO_MACOSX
+   CFStringRef cfstr = CFStringCreateWithCStringNoCopy(NULL, name,
+      kCFStringEncodingUTF8, kCFAllocatorNull);
+   if (cfstr) {
+      fptr = (VOID_FPTR)CFBundleGetFunctionPointerForName(opengl_bundle_ref, cfstr);
+      CFRelease(cfstr);
+   }
+#elif defined ALLEGRO_SDL
+   fptr = SDL_GL_GetProcAddress(name);
+#endif
+   if (fptr) {
+      ALLEGRO_DEBUG("%s successfully loaded (%p)\n", name, fptr);
+   }
+   return fptr;
+}
+
+
+
 /* Load the extension API addresses into the table.
  * Should only be done on context creation.
  */
@@ -296,17 +315,15 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #ifdef ALLEGRO_WINDOWS
 
-   #define AGL_API(type, name, args)                                 \
-      ext->name = (_ALLEGRO_gl##name##_t)wglGetProcAddress("gl" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" #name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
    #undef AGL_API
 
-   #define AGL_API(type, name, args)                                  \
-      ext->name = (_ALLEGRO_wgl##name##_t)wglGetProcAddress("wgl" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("wgl" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_wgl##name##_t)load_extension("wgl" #name);
 
       #include "allegro5/opengl/GLext/wgl_ext_api.h"
 
@@ -314,17 +331,15 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #elif defined ALLEGRO_UNIX
 
-   #define AGL_API(type, name, args)                                               \
-      ext->name = (_ALLEGRO_gl##name##_t)alXGetProcAddress((const GLubyte*)"gl" #name);  \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded (%p)\n", ext->name); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" #name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
    #undef AGL_API
 
-   #define AGL_API(type, name, args)                                               \
-      ext->name = (_ALLEGRO_glX##name##_t)alXGetProcAddress((const GLubyte*)"glX" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("glX" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_glX##name##_t)load_extension("glX" #name);
 
       #include "allegro5/opengl/GLext/glx_ext_api.h"
 
@@ -332,9 +347,17 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #elif defined ALLEGRO_MACOSX
 
-#define AGL_API(type, name, args)                                                                 \
-      ext->name = (_ALLEGRO_gl##name##_t)CFBundleGetFunctionPointerForName(opengl_bundle_ref, CFSTR("gl" # name)); \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded\n"); }
+#define AGL_API(type, name, args)                                              \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" # name);
+
+      #include "allegro5/opengl/GLext/gl_ext_api.h"
+
+   #undef AGL_API
+
+#elif defined ALLEGRO_SDL
+
+#define AGL_API(type, name, args)                                              \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" # name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
@@ -500,7 +523,6 @@ static bool _ogl_is_extension_supported(const char *extension,
 static bool _ogl_is_extension_with_version_supported(
    const char *extension, ALLEGRO_DISPLAY *disp, uint32_t ver)
 {
-   ALLEGRO_CONFIG *cfg;
    char const *value;
 
   /* For testing purposes, any OpenGL extension can be disable in
@@ -511,15 +533,12 @@ static bool _ogl_is_extension_with_version_supported(
     * GL_EXT_framebuffer_object=0
     * 
     */
-   cfg = al_get_system_config();
-   if (cfg) {
-      value = al_get_config_value(cfg,
-         "opengl_disabled_extensions", extension);
-      if (value) {
-         ALLEGRO_WARN("%s found in [opengl_disabled_extensions].\n",
-            extension);
-         return false;
-      }
+   value = al_get_config_value(al_get_system_config(),
+      "opengl_disabled_extensions", extension);
+   if (value) {
+      ALLEGRO_WARN("%s found in [opengl_disabled_extensions].\n",
+         extension);
+      return false;
    }
 
    /* If the extension is included in the OpenGL version, there is no
@@ -806,17 +825,6 @@ void _al_ogl_manage_extensions(ALLEGRO_DISPLAY *gl_disp)
 #undef AGL_EXT
 #endif
 
-    /* TODO: use these somewhere */
-#if 0
-   for (i = 0; i < 5; i++) {
-      __allegro_gl_texture_read_format[i] = -1;
-      __allegro_gl_texture_components[i] = GL_RGB;
-   }
-   __allegro_gl_texture_read_format[3] = GL_UNSIGNED_BYTE;
-   __allegro_gl_texture_read_format[4] = GL_UNSIGNED_BYTE;
-   __allegro_gl_texture_components[4] = GL_RGBA;
-#endif /* #if 0 */
-
    /* Get max texture size */
    glGetIntegerv(GL_MAX_TEXTURE_SIZE,
                  (GLint *) & gl_disp->ogl_extras->ogl_info.max_texture_size);
