@@ -27,13 +27,28 @@
 
 ALLEGRO_DEBUG_CHANNEL("opengl")
 
+/* FIXME: For some reason x86_64 Android crashes for me when calling
+ * glBlendColor - so adding this hack to disable it.
+ */
+#ifdef ALLEGRO_ANDROID
+#if defined(__x86_64__) || defined(__i686__)
+#define ALLEGRO_ANDROID_HACK_X86_64
+#endif
+#endif
+
 bool _al_opengl_set_blender(ALLEGRO_DISPLAY *ogl_disp)
 {
    int op, src_color, dst_color, op_alpha, src_alpha, dst_alpha;
-   const int blend_modes[8] = {
+   ALLEGRO_COLOR const_color;
+   const int blend_modes[10] = {
       GL_ZERO, GL_ONE, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
       GL_SRC_COLOR, GL_DST_COLOR, GL_ONE_MINUS_SRC_COLOR,
-      GL_ONE_MINUS_DST_COLOR
+      GL_ONE_MINUS_DST_COLOR,
+#if defined(ALLEGRO_CFG_OPENGLES2) || !defined(ALLEGRO_CFG_OPENGLES)
+      GL_CONSTANT_COLOR, GL_ONE_MINUS_CONSTANT_COLOR
+#else
+      GL_ONE, GL_ONE
+#endif
    };
    const int blend_equations[3] = {
       GL_FUNC_ADD, GL_FUNC_SUBTRACT, GL_FUNC_REVERSE_SUBTRACT
@@ -43,14 +58,28 @@ bool _al_opengl_set_blender(ALLEGRO_DISPLAY *ogl_disp)
 
    al_get_separate_blender(&op, &src_color, &dst_color,
       &op_alpha, &src_alpha, &dst_alpha);
+   const_color = al_get_blend_color();
    /* glBlendFuncSeparate was only included with OpenGL 1.4 */
-   /* (And not in OpenGL ES) */
 #if !defined ALLEGRO_CFG_OPENGLES
    if (ogl_disp->ogl_extras->ogl_info.version >= _ALLEGRO_OPENGL_VERSION_1_4) {
+#else
+   /* FIXME: At this time (09/2014) there are a lot of Android phones that
+    * don't support glBlendFuncSeparate even though they claim OpenGL ES 2.0
+    * support. Rather than not work on 20-25% of phones, we just don't support
+    * separate blending on Android for now.
+    */
+#ifdef ALLEGRO_ANDROID
+   if (false) {
 #else
    if (ogl_disp->ogl_extras->ogl_info.version >= _ALLEGRO_OPENGL_VERSION_2_0) {
+#endif
 #endif
       glEnable(GL_BLEND);
+#if defined(ALLEGRO_CFG_OPENGLES2) || !defined(ALLEGRO_CFG_OPENGLES)
+   #ifndef ALLEGRO_ANDROID_HACK_X86_64
+      glBlendColor(const_color.r, const_color.g, const_color.b, const_color.a);
+   #endif
+#endif
       glBlendFuncSeparate(blend_modes[src_color], blend_modes[dst_color],
          blend_modes[src_alpha], blend_modes[dst_alpha]);
       if (ogl_disp->ogl_extras->ogl_info.version >= _ALLEGRO_OPENGL_VERSION_2_0) {
@@ -65,6 +94,11 @@ bool _al_opengl_set_blender(ALLEGRO_DISPLAY *ogl_disp)
    else {
       if (src_color == src_alpha && dst_color == dst_alpha) {
          glEnable(GL_BLEND);
+#if defined(ALLEGRO_CFG_OPENGLES2) || !defined(ALLEGRO_CFG_OPENGLES)
+   #ifndef ALLEGRO_ANDROID_HACK_X86_64
+         glBlendColor(const_color.r, const_color.g, const_color.b, const_color.a);
+   #endif
+#endif
          glBlendFunc(blend_modes[src_color], blend_modes[dst_color]);
       }
       else {
@@ -192,13 +226,13 @@ static void ogl_clear_android_2_1_workaround(ALLEGRO_DISPLAY *d,
    };
    ALLEGRO_TRANSFORM bak1, bak2, t;
 
-   al_copy_transform(&bak1, &d->proj_transform);
+   al_copy_transform(&bak1, al_get_current_projection_transform());
    al_copy_transform(&bak2, al_get_current_transform());
 
    al_identity_transform(&t);
    al_orthographic_transform(&t, 0, 0, -1, d->w, d->h, 1);
 
-   al_set_projection_transform(d, &t);
+   al_use_projection_transform(&t);
    al_identity_transform(&t);
    al_use_transform(&t);
 
@@ -220,7 +254,7 @@ static void ogl_clear_android_2_1_workaround(ALLEGRO_DISPLAY *d,
    vert_ptr_off(d);
    color_ptr_off(d);
 
-   al_set_projection_transform(d, &bak1);
+   al_use_projection_transform(&bak1);
    al_use_transform(&bak2);
 }
 
@@ -329,6 +363,11 @@ static void ogl_flush_vertex_cache(ALLEGRO_DISPLAY *disp)
    if (disp->num_cache_vertices == 0)
       return;
 
+   if (!_al_opengl_set_blender(disp)) {
+      disp->num_cache_vertices = 0;
+      return;
+   }
+
    if (disp->flags & ALLEGRO_PROGRAMMABLE_PIPELINE) {
 #ifdef ALLEGRO_CFG_OPENGL_PROGRAMMABLE_PIPELINE
       if (disp->ogl_extras->varlocs.use_tex_loc >= 0) {
@@ -459,49 +498,32 @@ static void ogl_flush_vertex_cache(ALLEGRO_DISPLAY *disp)
 static void ogl_update_transformation(ALLEGRO_DISPLAY* disp,
    ALLEGRO_BITMAP *target)
 {
-   ALLEGRO_TRANSFORM tmp;
-   
-   al_copy_transform(&tmp, &target->transform);
-
-   if (target->parent) {
-      /* Sub-bitmaps have an additional offset. */
-      al_translate_transform(&tmp, target->xofs, target->yofs);
-   }
-
    if (disp->flags & ALLEGRO_PROGRAMMABLE_PIPELINE) {
 #ifdef ALLEGRO_CFG_SHADER_GLSL
       GLint loc = disp->ogl_extras->varlocs.projview_matrix_loc;
-      al_copy_transform(&disp->view_transform, &tmp);
+      ALLEGRO_TRANSFORM projview;
+      al_copy_transform(&projview, &target->transform);
+      al_compose_transform(&projview, &target->proj_transform);
+      al_copy_transform(&disp->projview_transform, &projview);
+
       if (disp->ogl_extras->program_object > 0 && loc >= 0) {
-         al_compose_transform(&tmp, &disp->proj_transform);
-         _al_glsl_set_projview_matrix(loc, &tmp);
+         _al_glsl_set_projview_matrix(loc, &disp->projview_transform);
       }
 #endif
-      return;
+   } else {
+      glMatrixMode(GL_PROJECTION);
+      glLoadMatrixf((float *)target->proj_transform.m);
+      glMatrixMode(GL_MODELVIEW);
+      glLoadMatrixf((float *)target->transform.m);
    }
 
-   glMatrixMode(GL_MODELVIEW);
-   glLoadMatrixf((float *)tmp.m);
-}
-
-static void ogl_set_projection(ALLEGRO_DISPLAY *d)
-{
-   if (d->flags & ALLEGRO_PROGRAMMABLE_PIPELINE) {
-#ifdef ALLEGRO_CFG_SHADER_GLSL
-      GLint loc = d->ogl_extras->varlocs.projview_matrix_loc;
-      if (d->ogl_extras->program_object > 0 && loc >= 0) {
-         ALLEGRO_TRANSFORM t;
-         al_copy_transform(&t, &d->view_transform);
-         al_compose_transform(&t, &d->proj_transform);
-         _al_glsl_set_projview_matrix(loc, &t);
-      }
-#endif
-      return;
+   if (target->parent) {
+      ALLEGRO_BITMAP_EXTRA_OPENGL *ogl_extra = target->parent->extra;
+      /* glViewport requires the bottom-left coordinate of the corner. */
+      glViewport(target->xofs, ogl_extra->true_h - (target->yofs + target->h), target->w, target->h);
+   } else {
+      glViewport(0, 0, target->w, target->h);
    }
-
-   glMatrixMode(GL_PROJECTION);
-   glLoadMatrixf((float *)d->proj_transform.m);
-   glMatrixMode(GL_MODELVIEW);
 }
 
 static void ogl_clear_depth_buffer(ALLEGRO_DISPLAY *display, float x)
@@ -530,7 +552,6 @@ void _al_ogl_add_drawing_functions(ALLEGRO_DISPLAY_INTERFACE *vt)
    vt->flush_vertex_cache = ogl_flush_vertex_cache;
    vt->prepare_vertex_cache = ogl_prepare_vertex_cache;
    vt->update_transformation = ogl_update_transformation;
-   vt->set_projection = ogl_set_projection;
 }
 
 /* vim: set sts=3 sw=3 et: */
